"""
Some possible attempts to save memory from the state-like simulator with checkpoint tricks (jax support only).
"""

import time
import sys
import logging

import numpy as np
import jax
import cotengra as ctg

logger = logging.getLogger("tensorcircuit")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)

sys.path.insert(0, "../")
sys.setrecursionlimit(10000)

import tensorcircuit as tc

optr = ctg.ReusableHyperOptimizer(
    methods=["greedy", "kahypar"],
    parallel=True,
    minimize="write",
    max_time=15,
    max_repeats=512,
    progbar=True,
)
tc.set_contractor("custom", optimizer=optr, preprocessing=True)
tc.set_dtype("complex64")
tc.set_backend("jax")


nwires, nlayers = 10, 36
sn = int(np.sqrt(nlayers))


def recursive_checkpoint(funs):
    if len(funs) == 1:
        return funs[0]
    elif len(funs) == 2:
        f1, f2 = funs
        return lambda s, param: f1(
            f2(s, param[: len(param) // 2]), param[len(param) // 2 :]
        )
    else:
        f1 = recursive_checkpoint(funs[len(funs) // 2 :])
        f2 = recursive_checkpoint(funs[: len(funs) // 2])
        return lambda s, param: f1(
            jax.checkpoint(f2)(s, param[: len(param) // 2]), param[len(param) // 2 :]
        )


# not suggest in general for recursive checkpoint: too slow for staging (compiling)

"""
test case:
def f(s, param):
    return s + param
fc = recursive_checkpoint([f for _ in range(100)])
print(fc(jnp.zeros([2]), jnp.array([[i, i] for i in range(100)])))
"""


@jax.checkpoint
@jax.jit
def zzxlayer(s, param):
    c = tc.Circuit(nwires, inputs=s)
    for i in range(0, nwires):
        c.exp1(
            i,
            (i + 1) % nwires,
            theta=param[0, i],
            unitary=tc.gates._zz_matrix,
        )
    for i in range(nwires):
        c.rx(i, theta=param[0, nwires + i])
    return c.state()


@jax.checkpoint
@jax.jit
def zzxsqrtlayer(s, param):
    for i in range(sn):
        s = zzxlayer(s, param[i : i + 1])
    return s


@jax.jit
def totallayer(s, param):
    for i in range(sn):
        s = zzxsqrtlayer(s, param[i * sn : (i + 1) * sn])
    return s


def vqe_forward(param):
    s = tc.backend.ones([2**nwires])
    s /= tc.backend.norm(s)
    s = totallayer(s, param)
    e = tc.expectation((tc.gates.x(), [1]), ket=s)
    return tc.backend.real(e)


def profile(tries=3):
    time0 = time.time()
    tc_vg = tc.backend.jit(tc.backend.value_and_grad(vqe_forward))
    param = tc.backend.cast(tc.backend.ones([nlayers, 2 * nwires]), "complex64")
    print(tc_vg(param))

    time1 = time.time()
    for _ in range(tries):
        print(tc_vg(param)[0])

    time2 = time.time()
    print(time1 - time0, (time2 - time1) / tries)


profile()



"""
Demonstration of the TFIM VQE on V100 with lager qubit number counts (100+).
"""

import time
import logging
import sys
import numpy as np

logger = logging.getLogger("tensorcircuit")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)

sys.setrecursionlimit(10000)

import tensorflow as tf
import cotengra as ctg

import tensorcircuit as tc
from tensorcircuit import keras

optr = ctg.ReusableHyperOptimizer(
    methods=["greedy", "kahypar"],
    parallel=True,
    minimize="flops",
    max_time=120,
    max_repeats=4096,
    progbar=True,
)
tc.set_contractor("custom", optimizer=optr, preprocessing=True)
# tc.set_contractor("custom_stateful", optimizer=oem.RandomGreedy, max_time=60, max_repeats=128, minimize="size")
tc.set_dtype("complex64")
tc.set_backend("tensorflow")
dtype = np.complex64

nwires, nlayers = 50, 7


def vqe_forward(param, structures):
    split_conf = {
        "max_singular_values": 2,
        "fixed_choice": 1,
    }
    structuresc = tc.backend.cast(structures, dtype="complex64")
    paramc = tc.backend.cast(param, dtype="complex64")
    c = tc.Circuit(nwires, split=split_conf)
    for i in range(nwires):
        c.H(i)
    for j in range(nlayers):
        for i in range(0, nwires - 1):
            c.exp1(
                i,
                (i + 1) % nwires,
                theta=paramc[2 * j, i],
                unitary=tc.gates._zz_matrix,
            )

        for i in range(nwires):
            c.rx(i, theta=paramc[2 * j + 1, i])

    obs = []
    for i in range(nwires):
        obs.append(
            [
                tc.gates.Gate(
                    sum(
                        [
                            structuresc[i, k] * g.tensor
                            for k, g in enumerate(tc.gates.pauli_gates)
                        ]
                    )
                ),
                (i,),
            ]
        )
    loss = c.expectation(*obs, reuse=False)
    return tc.backend.real(loss)


slist = []
for i in range(nwires):
    t = np.zeros(nwires)
    t[i] = 1
    slist.append(t)
for i in range(nwires):
    t = np.zeros(nwires)
    t[i] = 3
    t[(i + 1) % nwires] = 3
    slist.append(t)
structures = np.array(slist, dtype=np.int32)
structures = tc.backend.onehot(structures, num=4)
structures = tc.backend.reshape(structures, [-1, nwires, 4])
print(structures.shape)
time0 = time.time()

batch = 50
tc_vg = tf.function(
    tc.backend.vectorized_value_and_grad(vqe_forward, argnums=0, vectorized_argnums=1),
    input_signature=[
        tf.TensorSpec([2 * nlayers, nwires], tf.float32),
        tf.TensorSpec([batch, nwires, 4], tf.float32),
    ],
)
param = tf.Variable(tf.random.normal(stddev=0.1, shape=[2 * nlayers, nwires]))

print(tc_vg(param, structures[:batch]))

time1 = time.time()
print("staging time: ", time1 - time0)

try:
    keras.save_func(tc_vg, "./funcs/%s_%s_10_tfim" % (nwires, nlayers))
except ValueError as e:
    print(e)  # keras.save_func now has issues to be resolved


def train_step(param):
    vg_list = []
    for i in range(2):
        vg_list.append(tc_vg(param, structures[i * nwires : i * nwires + nwires]))
    loss = tc.backend.sum(vg_list[0][0] - vg_list[1][0])
    gr = vg_list[0][1] - vg_list[1][1]
    return loss, gr


if __name__ == "__main__":
    opt = tf.keras.optimizers.Adam(0.02)
    for j in range(5000):
        loss, gr = train_step(param)
        opt.apply_gradients([(gr, param)])
        if j % 20 == 0:
            print("loss", loss.numpy())



"""
Cross check the correctness of the density matrix simulator and the Monte Carlo trajectory state simulator.
"""
import os

os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
# cpu is fast for small scale circuit simulation
import sys

sys.path.insert(0, "../")

from tqdm import tqdm
import jax
import tensorcircuit as tc

tc.set_backend("jax")

n = 5
nlayer = 3
mctries = 100  # 100000

print(jax.devices())


def template(c):
    # dont jit me!
    for i in range(n):
        c.H(i)
    for i in range(n):
        c.rz(i, theta=tc.num_to_tensor(i))
    for _ in range(nlayer):
        for i in range(n - 1):
            c.cnot(i, i + 1)
        for i in range(n):
            c.rx(i, theta=tc.num_to_tensor(i))
        for i in range(n):
            c.apply_general_kraus(tc.channels.phasedampingchannel(0.15), i)
    return c.state()


@tc.backend.jit
def answer():
    c = tc.DMCircuit2(n)
    return template(c)


rho0 = answer()

print(rho0)


@tc.backend.jit
def f(key):
    if key is not None:
        tc.backend.set_random_state(key)
    c = tc.Circuit(n)
    return template(c)


key = jax.random.PRNGKey(42)
f(key).block_until_ready()  # build the graph

rho = 0.0

for i in tqdm(range(mctries)):
    key, subkey = jax.random.split(key)
    psi = f(subkey)  # [1, 2**n]
    rho += (
        1
        / mctries
        * tc.backend.reshape(psi, [-1, 1])
        @ tc.backend.conj(tc.backend.reshape(psi, [1, -1]))
    )

print(rho)
print("difference\n", tc.backend.abs(rho - rho0))
print("difference in total\n", tc.backend.sum(tc.backend.abs(rho - rho0)))
print("fidelity", tc.quantum.fidelity(rho, rho0))
print("trace distance", tc.quantum.trace_distance(rho, rho0))



"""
DQAS for QFT QEM circuit design, deprecated DQAS implementation
"""
import sys

sys.path.insert(0, "../")

from functools import partial
import cirq
import numpy as np
import tensorflow as tf
from tensorcircuit.applications.vags import qft_qem_vag
from tensorcircuit.applications.dqas import (
    set_op_pool,
    DQAS_search,
    verbose_output,
)


def main_3():
    qft_3 = partial(qft_qem_vag, n=3)

    set_op_pool([cirq.X, cirq.Y, cirq.Z, cirq.I, cirq.T, cirq.S])
    DQAS_search(
        qft_3,
        nnp_initial_value=tf.zeros([6, 6]),
        p=6,
        prethermal=0,
        batch=32,
        verbose=False,
        epochs=30,
    )


def main_4():
    qft_4 = partial(qft_qem_vag, n=4)

    set_op_pool(
        [
            cirq.I,
            cirq.X,
            cirq.Y,
            cirq.Z,
            cirq.H,
            cirq.rx(np.pi / 3),
            cirq.rx(np.pi * 2.0 / 3),
            cirq.rz(np.pi / 3),
            cirq.rz(np.pi * 2.0 / 3),
            cirq.S,
            cirq.T,
        ]
    )
    DQAS_search(
        qft_4,
        nnp_initial_value=tf.zeros([12, 11]),
        p=12,
        prethermal=0,
        batch=32,
        verbose=False,
        verbose_func=verbose_output,
        epochs=6,
    )


if __name__ == "__main__":
    # main_3()
    main_4()



"""
DQAS for GHZ state preparation circuit, deprecated DQAS implementation
"""
import sys

sys.path.insert(0, "../")
import numpy as np
import tensorflow as tf
import cirq

import tensorcircuit as tc
from tensorcircuit.applications.vags import double_qubits_initial, GHZ_vag, GHZ_vag_tfq
from tensorcircuit.applications.dqas import (
    set_op_pool,
    get_preset,
    DQAS_search,
)

tc.set_backend("tensorflow")


def main_tn():
    """
    DQAS with the tensorcircuit engine backend by TensorNetwork
    state preparation example

    :return:
    """
    # multi start may be necessary
    ghz_pool = [
        ("ry", 0),
        ("ry", 1),
        ("ry", 2),
        ("CNOT", 0, 1),
        ("CNOT", 1, 0),
        ("CNOT", 0, 2),
        ("CNOT", 2, 0),
        ("H", 0),
        ("H", 1),
        ("H", 2),
    ]
    set_op_pool(ghz_pool)
    c = len(ghz_pool)
    p = 4
    stp, nnp, _ = DQAS_search(
        GHZ_vag,
        p=p,
        batch=128,
        epochs=20,
        verbose=True,
        parallel_num=0,
        nnp_initial_value=np.zeros([p, c]),
        structure_opt=tf.keras.optimizers.Adam(learning_rate=0.15),
    )
    preset = get_preset(stp).numpy()
    GHZ_vag(None, nnp, preset, verbose=True)


def main_tfq():
    """
    DQAS with the tensorflow quantum engine.
    Unitary learning example.

    :return:
    """
    p = 4
    cset = [
        cirq.H(cirq.GridQubit(0, 0)),
        cirq.H(cirq.GridQubit(1, 0)),
        cirq.CNOT(cirq.GridQubit(0, 0), cirq.GridQubit(1, 0)),
        cirq.CNOT(cirq.GridQubit(1, 0), cirq.GridQubit(0, 0)),
        cirq.X(cirq.GridQubit(0, 0)),
        cirq.X(cirq.GridQubit(1, 0)),
    ]
    set_op_pool(cset)
    c = len(cset)
    stp, nnp, _ = DQAS_search(
        GHZ_vag_tfq,
        g=double_qubits_initial(),
        p=p,
        batch=16,
        epochs=5,
        verbose=False,
        parallel_num=0,
        nnp_initial_value=np.zeros([p, c]),
        structure_opt=tf.keras.optimizers.Adam(learning_rate=0.15),
    )
    preset = get_preset(stp).numpy()
    GHZ_vag_tfq(double_qubits_initial().send(None), nnp, preset, verbose=True)


if __name__ == "__main__":
    main_tfq()
    # main_tn()



"""
A plain QAOA optimization example with given graphs using networkx.
"""
import sys

sys.path.insert(0, "../")
import networkx as nx
import tensorflow as tf
import tensorcircuit as tc

K = tc.set_backend("tensorflow")

## 1. define the graph


def dict2graph(d):
    g = nx.to_networkx_graph(d)
    for e in g.edges:
        if not g[e[0]][e[1]].get("weight"):
            g[e[0]][e[1]]["weight"] = 1.0
    return g


# a graph instance

example_graph_dict = {
    0: {1: {"weight": 1.0}, 7: {"weight": 1.0}, 3: {"weight": 1.0}},
    1: {0: {"weight": 1.0}, 2: {"weight": 1.0}, 3: {"weight": 1.0}},
    2: {1: {"weight": 1.0}, 3: {"weight": 1.0}, 5: {"weight": 1.0}},
    4: {7: {"weight": 1.0}, 6: {"weight": 1.0}, 5: {"weight": 1.0}},
    7: {4: {"weight": 1.0}, 6: {"weight": 1.0}, 0: {"weight": 1.0}},
    3: {1: {"weight": 1.0}, 2: {"weight": 1.0}, 0: {"weight": 1.0}},
    6: {7: {"weight": 1.0}, 4: {"weight": 1.0}, 5: {"weight": 1.0}},
    5: {6: {"weight": 1.0}, 4: {"weight": 1.0}, 2: {"weight": 1.0}},
}

example_graph = dict2graph(example_graph_dict)

# 2. define the quantum ansatz

nlayers = 3


def QAOAansatz(gamma, beta, g=example_graph):
    n = len(g.nodes)
    c = tc.Circuit(n)
    for i in range(n):
        c.H(i)
    for j in range(nlayers):
        for e in g.edges:
            c.exp1(
                e[0],
                e[1],
                unitary=tc.gates._zz_matrix,
                theta=g[e[0]][e[1]].get("weight", 1.0) * gamma[j],
            )
        for i in range(n):
            c.rx(i, theta=beta[j])

    # calculate the loss function, max cut
    loss = 0.0
    for e in g.edges:
        loss += c.expectation_ps(z=[e[0], e[1]])

    return K.real(loss)


# 3. get compiled function for QAOA ansatz and its gradient

QAOA_vg = K.jit(K.value_and_grad(QAOAansatz, argnums=(0, 1)), static_argnums=2)


# 4. optimization loop

beta = K.implicit_randn(shape=[nlayers], stddev=0.1)
gamma = K.implicit_randn(shape=[nlayers], stddev=0.1)
opt = K.optimizer(tf.keras.optimizers.Adam(1e-2))

for i in range(100):
    loss, grads = QAOA_vg(gamma, beta, example_graph)
    print(K.numpy(loss))
    gamma, beta = opt.update(grads, [gamma, beta])  # gradient descent



"""
Old-fashioned DQAS code on QAOA ansatz design, now deprecated.
"""

# pylint: disable=wildcard-import

import sys

sys.path.insert(0, "../")

from collections import namedtuple
from pickle import dump
from matplotlib import pyplot as plt
import numpy as np
import tensorflow as tf
import tensorcircuit as tc
from tensorcircuit.applications.dqas import *
from tensorcircuit.applications.vags import *
from tensorcircuit.applications.layers import *
from tensorcircuit.applications.graphdata import regular_graph_generator

# qaoa_block_vag_energy = partial(qaoa_block_vag, f=(_identity, _neg))

tc.set_backend("tensorflow")


def main_layerwise_encoding():
    p = 5
    c = 7

    def noise():
        n = np.random.normal(loc=0.0, scale=0.002, size=[p, c])
        return tf.constant(n, dtype=tf.float32)

    def penalty_gradient(stp, nnp, lbd=0.15, lbd2=0.01):
        c = stp.shape[1]
        p = stp.shape[0]
        cost = tf.constant(
            [1.0, 1.0, 1.0, 1.0, 27 / 2 * 2.0, 27 / 2 * 2.0, 15 / 2.0], dtype=tf.float32
        )
        with tf.GradientTape() as t:
            prob = tf.math.exp(stp) / tf.tile(
                tf.math.reduce_sum(tf.math.exp(stp), axis=1)[:, tf.newaxis], [1, c]
            )
            penalty = 0.0
            for i in range(p - 1):
                penalty += lbd * tf.tensordot(prob[i], prob[i + 1], 1)
                penalty += lbd2 * tf.tensordot(cost, prob[i], [0, 0])
            penalty += lbd2 * tf.tensordot(cost, prob[p - 1], [0, 0])
            penalty_g = t.gradient(penalty, stp)
        return penalty_g

    learning_rate_sch = tf.keras.optimizers.schedules.InverseTimeDecay(0.12, 1, 0.01)

    DQAS_search(
        qaoa_vag_energy,
        g=regular_graph_generator(n=8, d=3),
        p=p,
        batch=8,
        prethermal=0,
        prethermal_preset=[0, 6, 1, 6, 1],
        epochs=3,
        parallel_num=2,
        pertubation_func=noise,
        nnp_initial_value=np.random.normal(loc=0.23, scale=0.06, size=[p, c]),
        stp_regularization=penalty_gradient,
        network_opt=tf.keras.optimizers.Adam(learning_rate=0.03),
        prethermal_opt=tf.keras.optimizers.Adam(learning_rate=0.04),
        structure_opt=tf.keras.optimizers.SGD(learning_rate=learning_rate_sch),
    )


result = namedtuple("result", ["epoch", "cand", "loss"])


def main_block_encoding():
    learning_rate_sch = tf.keras.optimizers.schedules.InverseTimeDecay(0.3, 1, 0.01)
    p = 6
    c = 8

    def record():
        return result(
            get_var("epoch"), get_var("cand_preset_repr"), get_var("avcost1").numpy()
        )

    def noise():
        # p = 6
        # c = 6
        n = np.random.normal(loc=0.0, scale=0.2, size=[2 * p, c])
        return tf.constant(n, dtype=tf.float32)

    stp, nnp, h = DQAS_search(
        qaoa_block_vag_energy,
        g=regular_graph_generator(n=8, d=3),
        batch=8,
        prethermal=0,
        prethermal_preset=[0, 6, 1, 6, 1],
        epochs=3,
        parallel_num=2,
        pertubation_func=noise,
        p=p,
        history_func=record,
        nnp_initial_value=np.random.normal(loc=0.23, scale=0.06, size=[2 * p, c]),
        network_opt=tf.keras.optimizers.Adam(learning_rate=0.04),
        prethermal_opt=tf.keras.optimizers.Adam(learning_rate=0.04),
        structure_opt=tf.keras.optimizers.SGD(learning_rate=learning_rate_sch),
    )

    with open("qaoa_block.result", "wb") as f:
        dump([stp.numpy(), nnp.numpy(), h], f)

    epochs = np.arange(len(h))
    data = np.array([r.loss for r in h])
    plt.plot(epochs, data)
    plt.xlabel("epoch")
    plt.ylabel("objective")
    plt.savefig("qaoa_block.pdf")


layer_pool = [Hlayer, rxlayer, rylayer, rzlayer, xxlayer, yylayer, zzlayer]
block_pool = [
    Hlayer,
    rx_zz_block,
    zz_ry_block,
    zz_rx_block,
    zz_rz_block,
    xx_rz_block,
    yy_rx_block,
    rx_rz_block,
]
set_op_pool(block_pool)

if __name__ == "__main__":
    # main_layerwise_encoding()
    main_block_encoding()



"""
Optimizing the parameterized circuit with progressively dense two-qubit gates,
as a potential approach to alleviate barren plateau.
"""
import sys

sys.path.insert(0, "../")
import tensorflow as tf
import numpy as np
import tensorcircuit as tc

K = tc.set_backend("tensorflow")

n = 10
nlayers = 3
g = tc.templates.graphs.Line1D(n)


def energy(params, structures, n, nlayers):
    structures = (K.sign(structures) + 1) / 2  # 0 or 1
    structures = K.cast(structures, params.dtype)
    c = tc.Circuit(n)
    for i in range(n):
        c.H(i)
    for j in range(nlayers):
        for i in range(n - 1):
            matrix = structures[j, i] * tc.gates._ii_matrix + (
                1.0 - structures[j, i]
            ) * (
                K.cos(params[2 * j + 1, i]) * tc.gates._ii_matrix
                + 1.0j * K.sin(params[2 * j + 1, i]) * tc.gates._zz_matrix
            )
            c.any(
                i,
                i + 1,
                unitary=matrix,
            )
        for i in range(n):
            c.rx(i, theta=params[2 * j, i])

    e = tc.templates.measurements.heisenberg_measurements(
        c, g, hzz=1, hxx=0, hyy=0, hx=-1, hy=0, hz=0
    )  # TFIM energy from expectation of circuit c defined on lattice given by g
    return e


vagf = K.jit(K.value_and_grad(energy, argnums=0), static_argnums=(2, 3))

params = np.random.uniform(size=[2 * nlayers, n])
structures = np.random.uniform(size=[nlayers, n])
params, structures = tc.array_to_tensor(params, structures)

opt = K.optimizer(tf.keras.optimizers.Adam(1e-2))

for i in range(300):
    if i % 20 == 0:
        structures -= 0.2 * K.ones([nlayers, n])
    # one can change the structures by tune the structure tensor value
    # this specifically equiv to add two qubit gates
    e, grads = vagf(params, structures, n, nlayers)
    params = opt.update(grads, params)
    print(K.numpy(e))



"""
Backend agnostic linear regression with gradient descent optimization:
a demonstration on most of core features and paradigm of tensorcircuit
"""

# this script shows how backend agnostic magic works, no code change is required to switch backend
# we also include jit, vmap and AD features in this pure classical example
# this demonstrates that tensorcircuit can serve as a solid unified ML library without any "quantumness"

import sys

sys.path.insert(0, "../")
import numpy as np
import tensorcircuit as tc

# (x, y) data preparation

nsamples = 100
k0 = 1.0
b0 = 0.0

xs0 = np.random.uniform(low=-1, high=1, size=[nsamples])
ys0 = k0 * xs0 + b0 + np.random.normal(scale=0.1, size=[nsamples])


def lr(xs, ys):
    """
    fully ML backend agnostic linear regression implementation
    """
    # construct the loss
    def loss_pointwise(x, y, param):
        k, b = param["k"], param["b"]
        yp = k * x + b
        return (yp - y) ** 2

    # we suppose this loss function only works for scalar, so that we can show the usage of ``vmap``

    loss_vmap = tc.backend.vmap(loss_pointwise, vectorized_argnums=(0, 1))

    # now we define the total loss for all data

    def loss(xs, ys, param):
        losses = loss_vmap(xs, ys, param)
        return tc.backend.sum(losses)

    # we get the jitted function to evaluate loss and its derivatives wrt. param

    loss_and_grad = tc.backend.jit(tc.backend.value_and_grad(loss, argnums=2))

    # setup initial values and optimizers

    weight = {"k": tc.backend.implicit_randn(), "b": tc.backend.implicit_randn()}

    if tc.backend.name == "tensorflow":
        import tensorflow as tf

        opt = tc.backend.optimizer(tf.keras.optimizers.Adam(1e-2))
    elif tc.backend.name == "jax":
        import optax

        opt = tc.backend.optimizer(optax.adam(1e-2))
    else:
        raise ValueError("Unsupported backend")

    # gradient descent optimization loop
    maxstep = 500
    for i in range(maxstep):
        loss, grad = loss_and_grad(xs, ys, weight)
        weight = opt.update(grad, weight)
        if i % 100 == 0 or i == maxstep - 1:
            print("optimized MSE loss after %s round: " % i, tc.backend.numpy(loss))

    return tc.backend.numpy(weight["k"]), tc.backend.numpy(weight["b"])


if __name__ == "__main__":

    for n in ["tensorflow", "jax"]:
        with tc.runtime_backend(n):  # runtime backend switch with context manager
            print("~~~~~~~~ using %s backend ~~~~~~~~" % n)
            xs_tensor, ys_tensor = tc.array_to_tensor(xs0, ys0, dtype="float32")
            print("predicted coefficient", lr(xs_tensor, ys_tensor))



import sys

sys.path.insert(0, "../")

import tensorcircuit as tc
from tensorcircuit.applications.dqas import (
    parallel_qaoa_train,
    single_generator,
    set_op_pool,
)
from tensorcircuit.applications.layers import *  # pylint: disable=wildcard-import
from tensorcircuit.applications.graphdata import get_graph

tc.set_backend("tensorflow")

set_op_pool([Hlayer, rxlayer, rylayer, rzlayer, xxlayer, yylayer, zzlayer])


if __name__ == "__main__":
    # old fashion example, prefer vvag for new generation software for this task
    parallel_qaoa_train(
        [0, 6, 1, 6, 1],
        single_generator(get_graph("8B")),
        tries=4,
        cores=2,
        batch=1,
        epochs=5,
        scale=0.8,
    )



"""
Demonstration of TFIM VQE with extra size in MPO formulation, with highly customizable scipy optimization interface
"""

import os

os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"
import time
import logging
import sys
import numpy as np
from scipy import optimize

logger = logging.getLogger("tensorcircuit")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)

sys.setrecursionlimit(10000)

import tensornetwork as tn
import cotengra as ctg

import tensorcircuit as tc

optc = ctg.ReusableHyperOptimizer(
    methods=["greedy", "kahypar"],
    parallel="ray",
    minimize="combo",
    max_time=6,
    max_repeats=4096,
    progbar=True,
)


def opt_reconf(inputs, output, size, **kws):
    tree = optc.search(inputs, output, size)
    tree_r = tree.subtree_reconfigure_forest(
        parallel="ray",
        progbar=True,
        num_trees=6,
        num_restarts=6,
        subtree_weight_what=("size",),
    )
    return tree_r.path()


tc.set_contractor("custom", optimizer=opt_reconf, preprocessing=True)
tc.set_dtype("complex64")
tc.set_backend("jax")
# jax backend is incompatible with keras.save

dtype = np.complex64

nwires, nlayers = 8, 2  # 600, 7


Jx = np.array([1.0 for _ in range(nwires - 1)])  # strength of xx interaction (OBC)
Bz = np.array([-1.0 for _ in range(nwires)])  # strength of transverse field
hamiltonian_mpo = tn.matrixproductstates.mpo.FiniteTFI(
    Jx, Bz, dtype=dtype
)  # matrix product operator
hamiltonian_mpo = tc.quantum.tn2qop(hamiltonian_mpo)


def vqe_forward(param):
    print("compiling")
    split_conf = {
        "max_singular_values": 2,
        "fixed_choice": 1,
    }
    c = tc.Circuit(nwires, split=split_conf)
    for i in range(nwires):
        c.H(i)
    for j in range(nlayers):
        for i in range(0, nwires - 1):
            c.exp1(
                i,
                (i + 1) % nwires,
                theta=param[4 * j, i],
                unitary=tc.gates._xx_matrix,
            )

        for i in range(nwires):
            c.rz(i, theta=param[4 * j + 1, i])
        for i in range(nwires):
            c.ry(i, theta=param[4 * j + 2, i])
        for i in range(nwires):
            c.rz(i, theta=param[4 * j + 3, i])
    return tc.templates.measurements.mpo_expectation(c, hamiltonian_mpo)


def scipy_optimize(
    f, x0, method, jac=None, tol=1e-9, maxiter=10000, step=1000, threshold=1e-6
):
    epoch = 0
    loss_prev = 0
    threshold = 1e-6
    count = 0
    times = []
    while epoch < maxiter:
        time0 = time.time()
        r = optimize.minimize(
            f, x0=x0, method=method, tol=tol, jac=jac, options={"maxiter": step}
        )
        time1 = time.time()
        times.append(time1 - time0)
        loss = r["fun"]
        epoch += step
        x0 = r["x"]
        print(epoch, loss)
        print(r["message"])
        if len(times) > 2:
            running_time = np.mean(times[1:]) / step
            staging_time = times[0] - running_time * step
            print("staging time: ", staging_time)
            print("running time: ", running_time)
        if abs(loss - loss_prev) < threshold:
            count += 1
        loss_prev = loss
        if count > 5 + int(2000 / step):
            break
    return loss, x0, epoch


if __name__ == "__main__":
    param = tc.backend.implicit_randn(stddev=0.1, shape=[4 * nlayers, nwires])
    vqe_ng = tc.interfaces.scipy_optimize_interface(
        vqe_forward, shape=[4 * nlayers, nwires], gradient=False, jit=True
    )
    vqe_g = tc.interfaces.scipy_optimize_interface(
        vqe_forward, shape=[4 * nlayers, nwires], gradient=True, jit=True
    )
    scipy_optimize(
        vqe_ng, tc.backend.numpy(param), method="COBYLA", jac=False, maxiter=50000
    )
    scipy_optimize(
        vqe_g,
        tc.backend.numpy(param),
        method="L-BFGS-B",
        jac=True,
        tol=1e-6,
        maxiter=50000,
    )
    # for BFGS, large tol is necessary, see
    # https://stackoverflow.com/questions/34663539/scipy-optimize-fmin-l-bfgs-b-returns-abnormal-termination-in-lnsrch



"""
Some chaotic properties calculations from the circuit state.
"""
from functools import partial
import sys

sys.path.insert(0, "../")
import tensorflow as tf
import tensorcircuit as tc

K = tc.set_backend("tensorflow")


# build the circuit


@partial(K.jit, static_argnums=(1, 2))
def get_state(params, n, nlayers, inputs=None):
    c = tc.Circuit(n, inputs=inputs)  # inputs is for input state of the circuit
    for i in range(nlayers):
        for j in range(n):
            c.ry(j, theta=params[i, j])
        for j in range(n):
            c.cnot(j, (j + 1) % n)
    # one can further customize the layout and gate type above
    return c.state()  # output wavefunction


params = K.implicit_randn([5, 10])

s = get_state(params, n=10, nlayers=5)

print(s)  # output state in vector form
rm = tc.quantum.reduced_density_matrix(s, cut=10 // 2)
print(tc.quantum.entropy(rm))
# entanglement
# for more quantum quantities functions, please refer to tc.quantum module


@partial(K.jit, static_argnums=(2, 3, 4))
def frame_potential(param1, param2, t, n, nlayers):
    s1 = get_state(param1, n, nlayers)
    s2 = get_state(param2, n, nlayers)
    inner = K.tensordot(K.conj(s1), s2, 1)
    return K.abs(inner) ** (2 * t)


# calculate several samples together using vmap

frame_potential_vmap = K.vmap(
    partial(frame_potential, t=1, n=10, nlayers=5), vectorized_argnums=(0, 1)
)

for _ in range(5):
    print(
        frame_potential_vmap(K.implicit_randn([3, 5, 10]), K.implicit_randn([3, 5, 10]))
    )
    # the first dimension is the batch

# get \partial \psi_i/ \partial \params_j (Jacobian)

jac_func = K.jacfwd(partial(get_state, n=10, nlayers=5))
print(jac_func(params))

# correlation


def get_zz(params, n, nlayers, inputs=None):
    s = get_state(params, n, nlayers, inputs)
    c = tc.Circuit(n, inputs=s)
    z1z2 = c.expectation([tc.gates.z(), [1]], [tc.gates.z(), [2]])
    return K.real(
        z1z2
    )  # one can also add several correlations together as energy estimation


# hessian matrix

h_func = K.hessian(partial(get_zz, n=10, nlayers=5))
# suggest jax backend for hessian and directly use `jax.hessian` may be better
print(h_func(params))

# optimization, suppose the energy we want to minimize is just z1z2 as above

vg_func = K.jit(K.value_and_grad(get_zz), static_argnums=(1, 2))
opt = K.optimizer(tf.keras.optimizers.Adam(1e-2))

for i in range(200):  # gradient descent
    energy, grads = vg_func(params, 10, 5)
    params = opt.update(grads, params)
    if i % 20 == 0:
        print(energy)  # see energy optimization dynamics



"""
A simple script to benchmark the approximation power of the MPS simulator.
"""
import sys

sys.path.insert(0, "../")

import tensorcircuit as tc

tc.set_backend("tensorflow")


def tfi_energy(c, n, j=1.0, h=-1.0):
    e = 0.0
    for i in range(n):
        e += h * c.expectation((tc.gates.x(), [i]))
    for i in range(n - 1):
        e += j * c.expectation((tc.gates.z(), [i]), (tc.gates.z(), [(i + 1) % n]))
    return e


def tfi_energy_mps(c, n, j=1.0, h=-1.0):
    e = 0.0
    for i in range(n):
        e += h * c.expectation_single_gate(tc.gates.x(), i)
    for i in range(n - 1):
        e += j * c.expectation_two_gates_product(
            tc.gates.z(), tc.gates.z(), i, (i + 1) % n
        )
    return e


def energy(param, mpsd=None):
    if mpsd is None:
        c = tc.Circuit(n)
    else:
        c = tc.MPSCircuit(n)
        c.set_truncation_rule(max_singular_values=mpsd)

    for i in range(n):
        c.H(i)
    for j in range(nlayers):
        for i in range(n - 1):
            c.exp1(
                i,
                (i + 1) % n,
                theta=param[2 * j, i],
                unitary=tc.gates._zz_matrix,
            )
        for i in range(n):
            c.rx(i, theta=param[2 * j + 1, i])
    if mpsd is None:
        e = tfi_energy(c, n)
    else:
        e = tfi_energy_mps(c, n)

    e = tc.backend.real(e)
    if mpsd is not None:
        fidelity = c._fidelity
    else:
        fidelity = None
    return e, c.state(), fidelity


n, nlayers = 15, 20
print("number of qubits: ", n)
print("number of layers: ", nlayers)

param = tc.backend.implicit_randu([2 * nlayers, n])
# param = tc.backend.ones([2 * nlayers, n])
# it turns out that the mps approximation power highly depends on the
# parameters, if we use ``param = tc.backend.ones``, the apprixmation ratio decays very fast
# At least, the estimated fidelity is a very good proxy metric for real fidelity
# as long as it is larger than 50%
e0, s0, _ = energy(param)
print(
    "entanglement: ",
    tc.backend.numpy(
        tc.quantum.entropy(tc.quantum.reduced_density_matrix(s0, cut=n // 2))
    ),
)

for mpsd in [2, 5, 10, 20, 50, 100]:
    e1, s1, f1 = energy(param, mpsd=mpsd)
    print("------------------------")
    print("bond dimension: ", mpsd)
    print(
        "exact energy: ",
        tc.backend.numpy(e0),
        "mps simulator energy: ",
        tc.backend.numpy(e1),
    )
    print(
        "energy relative error(%): ",
        tc.backend.numpy(tc.backend.abs((e1 - e0) / e0)) * 100,
    )
    print("estimated fidelity:", tc.backend.numpy(f1))
    print(
        "real fidelity:",
        tc.backend.numpy(
            tc.backend.abs(tc.backend.tensordot(tc.backend.conj(s1), s0, 1))
        ),
    )



"""
Quantum natural gradient descent demonstration with the TFIM VQE example.
"""

import sys
import time

sys.path.insert(0, "../")

import optax
import tensorflow as tf

import tensorcircuit as tc
from tensorcircuit import experimental

tc.set_backend("jax")  # tf, jax are both ok

n, nlayers = 7, 2
g = tc.templates.graphs.Line1D(n)


@tc.backend.jit
def state(params):
    params = tc.backend.reshape(params, [2 * nlayers, n])
    c = tc.Circuit(n)
    c = tc.templates.blocks.example_block(c, params, nlayers=nlayers)
    return c.state()


def energy(params):
    s = state(params)
    c = tc.Circuit(n, inputs=s)
    loss = tc.templates.measurements.heisenberg_measurements(
        c, g, hzz=1, hxx=0, hyy=0, hx=-1
    )
    return tc.backend.real(loss)


vags = tc.backend.jit(tc.backend.value_and_grad(energy))
lr = 1e-2
if tc.backend.name == "jax":
    oopt = optax.sgd(lr)
else:
    oopt = tf.keras.optimizers.SGD(lr)
opt = tc.backend.optimizer(oopt)

qng = tc.backend.jit(experimental.qng(state, mode="fwd"))
qngr = tc.backend.jit(experimental.qng(state, mode="rev"))
qng2 = tc.backend.jit(experimental.qng2(state, mode="fwd"))
qng2r = tc.backend.jit(experimental.qng2(state, mode="rev"))


def train_loop(params, i, qngf):
    qmetric = qngf(params)
    value, grad = vags(params)
    ngrad = tc.backend.solve(qmetric, grad, assume_a="sym")
    params = opt.update(ngrad, params)
    if i % 10 == 0:
        print(tc.backend.numpy(value))
    return params


def plain_train_loop(params, i):
    value, grad = vags(params)
    params = opt.update(grad, params)
    if i % 10 == 0:
        print(tc.backend.numpy(value))
    return params


def benchmark(f, *args):
    # params = tc.backend.implicit_randn([2 * nlayers * n])
    params = 0.1 * tc.backend.ones([2 * nlayers * n])
    params = tc.backend.cast(params, "float32")
    time0 = time.time()
    params = f(params, 0, *args)
    time1 = time.time()
    for i in range(100):
        params = f(params, i + 1, *args)
    time2 = time.time()

    print("staging time: ", time1 - time0, "running time: ", (time2 - time1) / 100)


if __name__ == "__main__":
    print("quantum natural gradient descent 1+f")
    benchmark(train_loop, qng)
    print("quantum natural gradient descent 1+r")
    benchmark(train_loop, qngr)
    print("quantum natural gradient descent 2+f")
    benchmark(train_loop, qng2)
    print("quantum natural gradient descent 2+r")
    benchmark(train_loop, qng2r)
    print("plain gradient descent")
    benchmark(plain_train_loop)



"""
Get molecule qubit format Hamiltonian from openfermion.
"""

import time
import tensorflow as tf
from openfermion.chem import MolecularData
from openfermion.transforms import (
    get_fermion_operator,
    binary_code_transform,
    get_fermion_operator,
    reorder,
    checksum_code,
)
from openfermion.utils import up_then_down
from openfermionpyscf import run_pyscf
from scipy import sparse

import tensorcircuit as tc


n = 4
multiplicity = 1
geometry = [("H", (0, 0, 0.95 * i)) for i in range(n)]
description = "H%s_0.95" % str(n)
basis = "sto-3g"
molecule = MolecularData(geometry, basis, multiplicity, description=description)
molecule = run_pyscf(molecule, run_mp2=True, run_cisd=True, run_ccsd=True, run_fci=True)
print(molecule.fci_energy, molecule.ccsd_energy)
fermion_hamiltonian = get_fermion_operator(molecule.get_molecular_hamiltonian())
b = binary_code_transform(
    reorder(fermion_hamiltonian, up_then_down), 2 * checksum_code(n, 1)
)
ls, w = tc.templates.chems.get_ps(b, 2 * n - 2)
time0 = time.time()
m = tc.quantum.PauliStringSum2COO_numpy(
    tf.constant(ls, dtype=tf.int64), tf.constant(w, dtype=tf.complex128)
)
time1 = time.time()
print(m)
print("tc takes time: ", time1 - time0)
tc.backend.sparse_dense_matmul(m, tc.backend.ones([2 ** (2 * n - 2)]))
time2 = time.time()
print("tc takes time for mvp: ", time2 - time1)
sparse.save_npz("./h-" + str(n) + "-chain.npz", m)
m2 = sparse.load_npz("./h-" + str(n) + "-chain.npz")
print(m2)

"""
from openfermion.linalg import LinearQubitOperator

# too slow to use
h = LinearQubitOperator(b)
ids = np.eye(2 ** h.n_qubits)
time3 = time.time()
m = h.dot(np.ones([2 ** h.n_qubits]))
time4 = time.time()
print("of takes time for mvp: ", time4 - time3)
"""



"""
Gradient evaluation comparison between qiskit and tensorcircut
"""

import time
import json
from functools import reduce
from operator import xor
import numpy as np


from qiskit.opflow import X, StateFn
from qiskit.circuit import QuantumCircuit, ParameterVector
from qiskit.opflow.gradients import Gradient, QFI, Hessian

import tensorcircuit as tc
from tensorcircuit import experimental


def benchmark(f, *args, trials=10):
    time0 = time.time()
    r = f(*args)
    time1 = time.time()
    for _ in range(trials):
        r = f(*args)
    time2 = time.time()
    if trials > 0:
        time21 = (time2 - time1) / trials
    else:
        time21 = 0
    ts = (time1 - time0, time21)
    print("staging time: %.6f s" % ts[0])
    if trials > 0:
        print("running time: %.6f s" % ts[1])
    return r, ts


def grad_qiskit(n, l, trials=2):
    hamiltonian = reduce(xor, [X for _ in range(n)])
    wavefunction = QuantumCircuit(n)
    params = ParameterVector("theta", length=3 * n * l)
    for j in range(l):
        for i in range(n - 1):
            wavefunction.cnot(i, i + 1)
        for i in range(n):
            wavefunction.rx(params[3 * n * j + i], i)
        for i in range(n):
            wavefunction.rz(params[3 * n * j + i + n], i)
        for i in range(n):
            wavefunction.rx(params[3 * n * j + i + 2 * n], i)

    # Define the expectation value corresponding to the energy
    op = ~StateFn(hamiltonian) @ StateFn(wavefunction)
    grad = Gradient().convert(operator=op, params=params)

    def get_grad_qiskit(values):
        value_dict = {params: values}
        grad_result = grad.assign_parameters(value_dict).eval()
        return grad_result

    return benchmark(get_grad_qiskit, np.ones([3 * n * l]), trials=trials)


def qfi_qiskit(n, l, trials=0):
    wavefunction = QuantumCircuit(n)
    params = ParameterVector("theta", length=3 * n * l)
    for j in range(l):
        for i in range(n - 1):
            wavefunction.cnot(i, i + 1)
        for i in range(n):
            wavefunction.rx(params[3 * n * j + i], i)
        for i in range(n):
            wavefunction.rz(params[3 * n * j + i + n], i)
        for i in range(n):
            wavefunction.rx(params[3 * n * j + i + 2 * n], i)

    nat_grad = QFI().convert(operator=StateFn(wavefunction), params=params)

    def get_qfi_qiskit(values):
        value_dict = {params: values}
        grad_result = nat_grad.assign_parameters(value_dict).eval()
        return grad_result

    return benchmark(get_qfi_qiskit, np.ones([3 * n * l]), trials=trials)


def hessian_qiskit(n, l, trials=0):
    hamiltonian = reduce(xor, [X for _ in range(n)])
    wavefunction = QuantumCircuit(n)
    params = ParameterVector("theta", length=3 * n * l)
    for j in range(l):
        for i in range(n - 1):
            wavefunction.cnot(i, i + 1)
        for i in range(n):
            wavefunction.rx(params[3 * n * j + i], i)
        for i in range(n):
            wavefunction.rz(params[3 * n * j + i + n], i)
        for i in range(n):
            wavefunction.rx(params[3 * n * j + i + 2 * n], i)

    # Define the expectation value corresponding to the energy
    op = ~StateFn(hamiltonian) @ StateFn(wavefunction)
    grad = Hessian().convert(operator=op, params=params)

    def get_hs_qiskit(values):
        value_dict = {params: values}
        grad_result = grad.assign_parameters(value_dict).eval()
        return grad_result

    return benchmark(get_hs_qiskit, np.ones([3 * n * l]), trials=trials)


def grad_tc(n, l, trials=10):
    def f(params):
        c = tc.Circuit(n)
        for j in range(l):
            for i in range(n - 1):
                c.cnot(i, i + 1)
            for i in range(n):
                c.rx(i, theta=params[3 * n * j + i])
            for i in range(n):
                c.rz(i, theta=params[3 * n * j + i + n])
            for i in range(n):
                c.rx(i, theta=params[3 * n * j + i + 2 * n])
        return tc.backend.real(c.expectation(*[[tc.gates.x(), [i]] for i in range(n)]))

    get_grad_tc = tc.backend.jit(tc.backend.grad(f))
    return benchmark(get_grad_tc, tc.backend.ones([3 * n * l], dtype="float32"))


def qfi_tc(n, l, trials=10):
    def s(params):
        c = tc.Circuit(n)
        for j in range(l):
            for i in range(n - 1):
                c.cnot(i, i + 1)
            for i in range(n):
                c.rx(i, theta=params[3 * n * j + i])
            for i in range(n):
                c.rz(i, theta=params[3 * n * j + i + n])
            for i in range(n):
                c.rx(i, theta=params[3 * n * j + i + 2 * n])
        return c.state()

    get_qfi_tc = tc.backend.jit(experimental.qng(s, mode="fwd"))
    return benchmark(get_qfi_tc, tc.backend.ones([3 * n * l], dtype="float32"))


def hessian_tc(n, l, trials=10):
    def f(params):
        c = tc.Circuit(n)
        for j in range(l):
            for i in range(n - 1):
                c.cnot(i, i + 1)
            for i in range(n):
                c.rx(i, theta=params[3 * n * j + i])
            for i in range(n):
                c.rz(i, theta=params[3 * n * j + i + n])
            for i in range(n):
                c.rx(i, theta=params[3 * n * j + i + 2 * n])
        return tc.backend.real(c.expectation(*[[tc.gates.x(), [i]] for i in range(n)]))

    get_hs_tc = tc.backend.jit(tc.backend.hessian(f))
    return benchmark(get_hs_tc, tc.backend.ones([3 * n * l], dtype="float32"))


results = {}

for n in [4, 6, 8, 10, 12]:
    for l in [2, 4, 6]:
        _, ts = grad_qiskit(n, l)
        results[str(n) + "-" + str(l) + "-" + "grad" + "-qiskit"] = ts[1]
        _, ts = qfi_qiskit(n, l)
        results[str(n) + "-" + str(l) + "-" + "qfi" + "-qiskit"] = ts[0]
        _, ts = hessian_qiskit(n, l)
        results[str(n) + "-" + str(l) + "-" + "hs" + "-qiskit"] = ts[0]
        with tc.runtime_backend("tensorflow"):
            _, ts = grad_tc(n, l)
            results[str(n) + "-" + str(l) + "-" + "grad" + "-tc-tf"] = ts
            _, ts = qfi_tc(n, l)
            results[str(n) + "-" + str(l) + "-" + "qfi" + "-tc-tf"] = ts
            _, ts = hessian_tc(n, l)
            results[str(n) + "-" + str(l) + "-" + "hs" + "-tc-tf"] = ts
        with tc.runtime_backend("jax"):
            _, ts = grad_tc(n, l)
            results[str(n) + "-" + str(l) + "-" + "grad" + "-tc-jax"] = ts
            _, ts = qfi_tc(n, l)
            results[str(n) + "-" + str(l) + "-" + "qfi" + "-tc-jax"] = ts
            _, ts = hessian_tc(n, l)
            results[str(n) + "-" + str(l) + "-" + "hs" + "-tc-jax"] = ts

print(results)

with open("gradient_results.data", "w") as f:
    json.dump(results, f)

with open("gradient_results.data", "r") as f:
    print(json.load(f))



"""
Demonstration of TFIM VQE with extra size in MPO formulation
"""

import time
import logging
import sys
import numpy as np

logger = logging.getLogger("tensorcircuit")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)

sys.setrecursionlimit(10000)

import tensorflow as tf
import tensornetwork as tn
import cotengra as ctg
import optax

import tensorcircuit as tc

opt = ctg.ReusableHyperOptimizer(
    methods=["greedy", "kahypar"],
    parallel="ray",
    minimize="combo",
    max_time=360,
    max_repeats=4096,
    progbar=True,
)


def opt_reconf(inputs, output, size, **kws):
    tree = opt.search(inputs, output, size)
    tree_r = tree.subtree_reconfigure_forest(
        parallel="ray",
        progbar=True,
        num_trees=20,
        num_restarts=20,
        subtree_weight_what=("size",),
    )
    return tree_r.path()


tc.set_contractor("custom", optimizer=opt_reconf, preprocessing=True)
tc.set_dtype("complex64")
tc.set_backend("tensorflow")
# jax backend is incompatible with keras.save

dtype = np.complex64

nwires, nlayers = 150, 7  # 600, 7


Jx = np.array([1.0 for _ in range(nwires - 1)])  # strength of xx interaction (OBC)
Bz = np.array([-1.0 for _ in range(nwires)])  # strength of transverse field
hamiltonian_mpo = tn.matrixproductstates.mpo.FiniteTFI(
    Jx, Bz, dtype=dtype
)  # matrix product operator
hamiltonian_mpo = tc.quantum.tn2qop(hamiltonian_mpo)


def vqe_forward(param):
    print("compiling")
    split_conf = {
        "max_singular_values": 2,
        "fixed_choice": 1,
    }
    c = tc.Circuit(nwires, split=split_conf)
    for i in range(nwires):
        c.H(i)
    for j in range(nlayers):
        for i in range(0, nwires - 1):
            c.exp1(
                i,
                (i + 1) % nwires,
                theta=param[4 * j, i],
                unitary=tc.gates._xx_matrix,
            )

        for i in range(nwires):
            c.rz(i, theta=param[4 * j + 1, i])
        for i in range(nwires):
            c.ry(i, theta=param[4 * j + 2, i])
        for i in range(nwires):
            c.rz(i, theta=param[4 * j + 3, i])
    return tc.templates.measurements.mpo_expectation(c, hamiltonian_mpo)


if __name__ == "__main__":
    refresh = False

    time0 = time.time()
    if refresh:
        tc_vg = tf.function(
            tc.backend.value_and_grad(vqe_forward),
            input_signature=[tf.TensorSpec([4 * nlayers, nwires], tf.float32)],
        )
        tc.keras.save_func(tc_vg, "./funcs/%s_%s_tfim_mpo" % (nwires, nlayers))
        time1 = time.time()
        print("staging time: ", time1 - time0)

    tc_vg_loaded = tc.keras.load_func("./funcs/%s_%s_tfim_mpo" % (nwires, nlayers))

    lr1 = 0.008
    lr2 = 0.06
    steps = 2000
    switch = 400
    debug_steps = 20

    if tc.backend.name == "jax":
        opt = tc.backend.optimizer(optax.adam(lr1))
        opt2 = tc.backend.optimizer(optax.sgd(lr2))
    else:
        opt = tc.backend.optimizer(tf.keras.optimizers.Adam(lr1))
        opt2 = tc.backend.optimizer(tf.keras.optimizers.SGD(lr2))

    times = []
    param = tc.backend.implicit_randn(stddev=0.1, shape=[4 * nlayers, nwires])

    for j in range(steps):
        loss, gr = tc_vg_loaded(param)
        if j < switch:
            param = opt.update(gr, param)
        else:
            if j == switch:
                print("switching the optimizer")
            param = opt2.update(gr, param)
        if j % debug_steps == 0 or j == steps - 1:
            times.append(time.time())
            print("loss", tc.backend.numpy(loss))
            if j > 0:
                print("running time:", (times[-1] - times[0]) / j)


"""
# Baseline code: obtained from DMRG using quimb

import quimb

h = quimb.tensor.tensor_gen.MPO_ham_ising(nwires, 4, 2, cyclic=False)
dmrg = quimb.tensor.tensor_dmrg.DMRG2(m, bond_dims=[10, 20, 100, 100, 200], cutoffs=1e-10)
dmrg.solve(tol=1e-9, verbosity=1) # may require repetition of this API
"""



"""
H6 molecule VQNHE with code from tc.application
"""
import sys

sys.path.insert(0, "../")
import numpy as np
import tensorcircuit as tc
from tensorcircuit.applications.vqes import VQNHE, JointSchedule, construct_matrix_v3

tc.set_backend("tensorflow")
tc.set_dtype("complex128")

h6h = np.load("./h6_hamiltonian.npy")  # reported in 0.99 A
hamiltonian = construct_matrix_v3(h6h.tolist())


vqeinstance = VQNHE(
    10,
    hamiltonian,
    {"width": 16, "stddev": 0.001, "choose": "complex-rbm"},  # model parameter
    {"filled_qubit": [0, 1, 3, 4, 5, 6, 8, 9], "epochs": 2},  # circuit parameter
    shortcut=True,  # enable shortcut for full Hamiltonian matrix evaluation
)
# 1110011100

rs = vqeinstance.multi_training(
    tries=2,  # 10
    maxiter=500,  # 10000
    threshold=0.5e-8,
    optq=JointSchedule(200, 0.01, 800, 0.002, 800),
    optc=JointSchedule(200, 0.0006, 10000, 0.008, 5000),
    onlyq=0,
    debug=200,
    checkpoints=[(900, -3.18), (2600, -3.19), (4500, -3.2)],
)
print(rs)



"""
Time comparison for different evaluation approach on molecule VQE
"""

import os

os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"
# one need this for jax+gpu combination in some cases
import time
import numpy as np
from openfermion.chem import MolecularData
from openfermion.transforms import (
    get_fermion_operator,
    binary_code_transform,
    checksum_code,
    reorder,
)
from openfermion.chem import geometry_from_pubchem
from openfermion.utils import up_then_down
from openfermionpyscf import run_pyscf

import tensorcircuit as tc

K = tc.set_backend("tensorflow")


n = 12
nlayers = 2
multiplicity = 1
basis = "sto-3g"
# 14 spin orbitals for H2O
geometry = geometry_from_pubchem("h2o")
description = "h2o"
molecule = MolecularData(geometry, basis, multiplicity, description=description)
molecule = run_pyscf(molecule, run_mp2=True, run_cisd=True, run_ccsd=True, run_fci=True)
mh = molecule.get_molecular_hamiltonian()
fh = get_fermion_operator(mh)
b = binary_code_transform(reorder(fh, up_then_down), 2 * checksum_code(7, 1))
lsb, wb = tc.templates.chems.get_ps(b, 12)
print("%s terms in H2O qubit Hamiltonian" % len(wb))
mb = tc.quantum.PauliStringSum2COO_numpy(lsb, wb)
mbd = mb.todense()
mb = K.coo_sparse_matrix(
    np.transpose(np.stack([mb.row, mb.col])), mb.data, shape=(2**n, 2**n)
)
mbd = tc.array_to_tensor(mbd)


def ansatz(param):
    c = tc.Circuit(n)
    for i in [0, 1, 2, 3, 4, 6, 7, 8, 9, 10]:
        c.X(i)
    for j in range(nlayers):
        for i in range(n - 1):
            c.cz(i, i + 1)
        for i in range(n):
            c.rx(i, theta=param[j, i])
    return c


def benchmark(vqef, tries=3):
    if tries < 0:
        vagf = K.value_and_grad(vqef)
    else:
        vagf = K.jit(K.value_and_grad(vqef))
    time0 = time.time()
    v, g = vagf(K.zeros([nlayers, n]))
    time1 = time.time()
    for _ in range(tries):
        v, g = vagf(K.zeros([nlayers, n]))
    time2 = time.time()
    print("staging time: ", time1 - time0)
    if tries > 0:
        print("running time: ", (time2 - time1) / tries)
    return (v, g), (time1 - time0, (time2 - time1) / tries)


# 1. plain Pauli sum


def vqe1(param):
    c = ansatz(param)
    loss = 0.0
    for ps, w in zip(lsb, wb):
        obs = []
        for i, p in enumerate(ps):
            if p == 1:
                obs.append([tc.gates.x(), [i]])
            elif p == 2:
                obs.append([tc.gates.y(), [i]])
            elif p == 3:
                obs.append([tc.gates.z(), [i]])

        loss += w * c.expectation(*obs)
    return K.real(loss)


# 2. vmap the Pauli sum


def measurement(s, structure):
    c = tc.Circuit(n, inputs=s)
    return tc.templates.measurements.parameterized_measurements(
        c, structure, onehot=True
    )


measurement = K.jit(K.vmap(measurement, vectorized_argnums=1))
structures = tc.array_to_tensor(lsb)
weights = tc.array_to_tensor(wb, dtype="float32")


def vqe2(param):
    c = ansatz(param)
    s = c.state()
    ms = measurement(s, structures)
    return K.sum(ms * weights)


# 3. dense matrix


def vqe3(param):
    c = ansatz(param)
    return tc.templates.measurements.operator_expectation(c, mbd)


# 4. sparse matrix


def vqe4(param):
    c = ansatz(param)
    return tc.templates.measurements.operator_expectation(c, mb)


# 5. mpo (ommited, since it is not that applicable for molecule/long range Hamiltonian
# due to large bond dimension)


if __name__ == "__main__":
    r0 = None
    des = [
        "plain Pauli sum",
        "vmap Pauli sum",
        "dense Hamiltonian matrix",
        "sparse Hamiltonian matrix",
    ]
    vqef = [vqe1, vqe2, vqe3, vqe4]
    tries = [-1, 2, 5, 5]
    for i in range(4):
        print(des[i])
        r1, _ = benchmark(vqef[i], tries=tries[i])
        # plain approach takes too long to jit
        if r0 is not None:
            np.testing.assert_allclose(r0[0], r1[0], atol=1e-5)
            np.testing.assert_allclose(r0[1], r1[1], atol=1e-5)
        r0 = r1
        print("------------------")



"""
Calculate the molecule dissociation curve using VQNHE.
"""
from functools import partial
import sys

sys.path.insert(0, "../")

import numpy as np
import tensorflow as tf

import tensorcircuit as tc
from tensorcircuit.applications.vqes import VQNHE, JointSchedule

tc.set_backend("tensorflow")
tc.set_dtype("complex128")


def initial_param(t, last=None, lastlast=None):
    if ((t % 3 == 1) and last) or ((t % 3 == 2) and lastlast):
        if t % 3 == 2:
            last = lastlast
        qw = last[-1]
        qw = tf.Variable(
            qw.numpy() + np.random.uniform(low=-0.1, high=0.1, size=qw.numpy().shape)
        )
        cw = last[-2]
        for i, b in enumerate(cw):
            cw[i] = b + np.random.uniform(low=-0.1, high=0.1, size=t.shape)
        return {"c": cw, "q": qw}

    return {}


def adiabatic_range(hm, history):
    if len(history) > 0:
        last = sorted(
            [
                (
                    r["energy"],
                    r["quantum_energy"],
                    r["model_weights"],
                    r["circuit_weights"],
                )
                for r in history[-1]
            ],
            key=lambda s: s[0],
        )[0]
    else:
        last = None
    if len(history) > 1:
        lastlast = sorted(
            [
                (
                    r["energy"],
                    r["quantum_energy"],
                    r["model_weights"],
                    r["circuit_weights"],
                )
                for r in history[-1]
            ],
            key=lambda s: s[0],
        )[0]
    else:
        lastlast = None
    print("begin calculation on new")
    vqeinstance = VQNHE(
        4,
        hm,
        {"max_value": 5, "init_value": 1.0, "min_value": 0.1},
        {"filled_qubit": [0]},
    )

    def learn_q():
        return JointSchedule(180, 0.009, 800, 0.001, 800)

    def learn_c():
        return JointSchedule(160, 0.002, 10000, 0.2, 1500)

    rs = vqeinstance.multi_training(
        tries=2,
        maxiter=150,  # 10000
        threshold=0.2 * 1e-8,
        optq=learn_q,  # JointSchedule(2800, 0.009, 800, 0.002, 100),
        optc=learn_c,
        initialization_func=partial(initial_param, last=last, lastlast=lastlast),
    )
    print(
        sorted(
            [(r["energy"], r["quantum_energy"], r["iterations"]) for r in rs],
            key=lambda s: s[0],
        )
    )
    return rs


if __name__ == "__main__":
    history = []
    lihh = np.load("data_file")
    for h in lihh[3:6]:
        history.append(adiabatic_range(h.tolist(), history))
    print(history)
    # vqeinstance = VQNHE(
    #     4,
    #     lihh,
    #     {"max_value": 5, "init_value": 1.0, "min_value": 0.1},
    #     {"filled_qubit": [0]},
    # )

    # def learn_q():
    #     return JointSchedule(180, 0.009, 800, 0.001, 800)

    # def learn_c():
    #     return JointSchedule(160, 0.002, 10000, 0.2, 1500)

    # rs = vqeinstance.multi_training(
    #     tries=10,
    #     maxiter=15000,
    #     threshold=0.2 * 1e-8,
    #     learn_q=learn_q,  # JointSchedule(2800, 0.009, 800, 0.002, 100),
    #     learn_c=learn_c,
    # )
    # print(rs)
    # print(
    #     sorted(
    #         [(r["energy"], r["quantum_energy"], r["iterations"]) for r in rs],
    #         key=lambda s: s[0],
    #     )
    # )



"""
benchmark sparse hamiltonian building
"""
import time
import numpy as np
import quimb
import tensorcircuit as tc


nwires = 20
# tc outperforms quimb in large nwires


# 1. tc approach for TFIM

g = tc.templates.graphs.Line1D(nwires, pbc=False)
time0 = time.time()
h1 = tc.quantum.heisenberg_hamiltonian(
    g, hzz=1, hxx=0, hyy=0, hz=0, hx=-1, hy=0, sparse=True, numpy=True
)
time1 = time.time()

print("tc time: ", time1 - time0)

# 2. quimb approach for TFIM

builder = quimb.tensor.tensor_gen.SpinHam()
# spin operator instead of Pauli matrix
builder += 4, "Z", "Z"
builder += -2, "X"
time0 = time.time()
h2 = builder.build_sparse(nwires)
h2 = h2.tocoo()
time1 = time.time()

print("quimb time: ", time1 - time0)


def assert_equal(h1, h2):
    np.testing.assert_allclose(h1.row, h2.row, atol=1e-5)
    np.testing.assert_allclose(h1.col, h2.col, atol=1e-5)
    np.testing.assert_allclose(h1.data, h2.data, atol=1e-5)


assert_equal(h1, h2)



"""
Optimization for performance comparison for different densities of two-qubit gates
(random layouts averaged).
"""
import sys

sys.path.insert(0, "../")
import tensorflow as tf
import numpy as np
import cotengra as ctg
import tensorcircuit as tc

K = tc.set_backend("tensorflow")
optr = ctg.ReusableHyperOptimizer(
    methods=["greedy"],
    parallel=True,
    minimize="flops",
    max_time=30,
    max_repeats=512,
    progbar=True,
)
tc.set_contractor("custom", optimizer=optr, preprocessing=True)

eye4 = K.eye(4)
cnot = tc.array_to_tensor(tc.gates._cnot_matrix)


def energy_p(params, p, seed, n, nlayers):
    g = tc.templates.graphs.Line1D(n)
    c = tc.Circuit(n)
    for i in range(n):
        c.H(i)
    for i in range(nlayers):
        for k in range(n):
            c.ry(k, theta=params[2 * i, k])
            c.rz(k, theta=params[2 * i + 1, k])
            c.ry(k, theta=params[2 * i + 2, k])
        for k in range(n // 2):  # alternating entangler with probability
            c.unitary_kraus(
                [eye4, cnot],
                2 * k + (i % 2),
                (2 * k + (i % 2) + 1) % n,
                prob=[1 - p, p],
                status=seed[i, k],
            )

    e = tc.templates.measurements.heisenberg_measurements(
        c, g, hzz=1, hxx=0, hyy=0, hx=-1, hy=0, hz=0
    )  # TFIM energy from expectation of circuit c defined on lattice given by g
    return e


vagf = K.jit(
    K.vectorized_value_and_grad(energy_p, argnums=0, vectorized_argnums=(0, 2)),
    static_argnums=(3, 4),
)

energy_list = []


if __name__ == "__main__":

    n = 12
    nlayers = 12
    nsteps = 250
    sample = 10
    debug = True

    for a in np.arange(0.1, 1, 0.1):
        energy_sublist = []
        params = K.implicit_randn(shape=[sample, 3 * nlayers, n])
        seeds = K.implicit_randu(shape=[sample, nlayers, n // 2])
        opt = tc.backend.optimizer(tf.keras.optimizers.Adam(2e-3))
        for i in range(nsteps):
            p = (n * nlayers) ** (a - 1)
            p = tc.array_to_tensor(p, dtype="float32")
            e, grads = vagf(params, p, seeds, n, nlayers)
            params = opt.update(grads, params)
            if i % 50 == 0 and debug:
                print(a, i, e)
        energy_list.append(K.numpy(e))

    print(energy_list)



"""
Variational wavefunctions based on variational circuits and its dynamics.
"""

# Variational Quantum Algorithm for Quantum Dynamics
# Ref: PRL 125, 010501 (2020)

import sys
import time

sys.path.insert(0, "../")

import numpy as np
import tensorcircuit as tc

tc.set_backend("jax")  # can set tensorflow backend or jax backend
tc.set_dtype("complex64")
# the default precision is complex64, can change to complex128 for double precision


def variational_wfn(theta, psi0):
    theta = tc.backend.reshape(theta, [l, N, 2])
    c = tc.Circuit(N, inputs=psi0)
    for i in range(0, l):
        for j in range(N - 1):
            c.exp1(j, j + 1, theta=theta[i, j, 0], unitary=tc.gates._zz_matrix)
        for j in range(N):
            c.rx(j, theta=theta[i, j, 1])

    return c.state()


ppsioverptheta = tc.backend.jit(tc.backend.jacfwd(variational_wfn, argnums=0))
# compute \partial psi /\partial theta, i.e. jacobian of wfn


def ij(i, j):
    """
    Inner product
    """
    return tc.backend.tensordot(tc.backend.conj(i), j, 1)


@tc.backend.jit
def lhs_matrix(theta, psi0):
    vij = tc.backend.vmap(ij, vectorized_argnums=0)
    vvij = tc.backend.vmap(vij, vectorized_argnums=1)
    jacobian = ppsioverptheta(theta, psi0=psi0)
    jacobian = tc.backend.transpose(jacobian)
    fim = vvij(jacobian, jacobian)
    fim = tc.backend.real(fim)
    return fim


@tc.backend.jit
def rhs_vector(theta, psi0):
    def energy(theta, psi0):
        w = variational_wfn(theta, psi0)
        wl = tc.backend.stop_gradient(w)
        wl = tc.backend.conj(wl)
        wr = w
        wl = tc.backend.reshape(wl, [1, -1])
        wr = tc.backend.reshape(wr, [-1, 1])
        e = wl @ h @ wr
        return tc.backend.real(e)[0, 0]

    eg = tc.backend.grad(energy, argnums=0)
    rhs = eg(theta, psi0)
    rhs = tc.backend.imag(rhs)
    return rhs


@tc.backend.jit
def update(theta, lhs, rhs, tau):
    # protection
    eps = 1e-4
    lhs += eps * tc.backend.eye(l * N * 2, dtype=lhs.dtype)
    return (
        tc.backend.cast(
            tau * tc.backend.solve(lhs, rhs, assume_a="sym"), dtype=theta.dtype
        )
        + theta
    )


if __name__ == "__main__":
    N = 10
    l = 5
    tau = 0.005
    steps = 200

    g = tc.templates.graphs.Line1D(N, pbc=False)
    h = tc.quantum.heisenberg_hamiltonian(
        g, hzz=1, hyy=0, hxx=0, hz=0, hx=1, hy=0, sparse=False
    )
    # TFIM Hamiltonian defined on lattice graph g (1D OBC chain)
    h = tc.array_to_tensor(h)

    psi0 = np.zeros(2**N)
    psi0[0] = 1.0
    psi0 = tc.array_to_tensor(psi0)

    theta = np.zeros([l * N * 2])
    theta = tc.array_to_tensor(theta)

    time0 = time.time()

    for n in range(steps):
        psi = variational_wfn(theta, psi0)
        lhs = lhs_matrix(theta, psi0)
        rhs = rhs_vector(theta, psi0)
        theta = update(theta, lhs, rhs, tau)
        if n % 10 == 0:
            time1 = time.time()
            print(time1 - time0)
            time0 = time1
            psi_exact = tc.backend.expm(-1j * h * n * tau) @ tc.backend.reshape(
                psi0, [-1, 1]
            )
            psi_exact = tc.backend.reshape(psi_exact, [-1])
            print(
                "time: %.2f" % (n * tau),
                "exact:",
                tc.expectation([tc.gates.z(), [0]], ket=psi_exact),
                "variational:",
                tc.expectation([tc.gates.z(), [0]], ket=psi),
            )



"""
Boost the Monte Carlo noise simulation (specifically the staging time)
on general error with circuit layerwise slicing
"""

import time
import sys

sys.path.insert(0, "../")
import tensorcircuit as tc

tc.set_backend("jax")

n = 3  # 10
nlayer = 2  # 4


def f1(key, param, n, nlayer):
    if key is not None:
        tc.backend.set_random_state(key)
    c = tc.Circuit(n)
    for i in range(n):
        c.H(i)
    for j in range(nlayer):
        for i in range(n - 1):
            c.cnot(i, i + 1)
            c.apply_general_kraus(tc.channels.phasedampingchannel(0.15), i)
            c.apply_general_kraus(tc.channels.phasedampingchannel(0.15), i + 1)
        for i in range(n):
            c.rx(i, theta=param[j, i])
    return tc.backend.real(c.expectation((tc.gates.z(), [int(n / 2)])))


def templatecnot(s, param, i):
    c = tc.Circuit(n, inputs=s)
    c.cnot(i, i + 1)
    return c.state()


def templatenoise(key, s, param, i):
    c = tc.Circuit(n, inputs=s)
    status = tc.backend.stateful_randu(key)[0]
    c.apply_general_kraus(tc.channels.phasedampingchannel(0.15), i, status=status)
    return c.state()


def templaterz(s, param, j):
    c = tc.Circuit(n, inputs=s)
    for i in range(n):
        c.rx(i, theta=param[j, i])
    return c.state()


def f2(key, param, n, nlayer):
    c = tc.Circuit(n)
    for i in range(n):
        c.H(i)
    s = c.state()
    for j in range(nlayer):
        for i in range(n - 1):
            s = templatecnot(s, param, i)
            key, subkey = tc.backend.random_split(key)
            s = templatenoise(subkey, s, param, i)
            key, subkey = tc.backend.random_split(key)
            s = templatenoise(subkey, s, param, i + 1)
        s = templaterz(s, param, j)
    return tc.backend.real(tc.expectation((tc.gates.z(), [int(n / 2)]), ket=s))


vagf1 = tc.backend.jit(tc.backend.value_and_grad(f1, argnums=1), static_argnums=(2, 3))
vagf2 = tc.backend.jit(tc.backend.value_and_grad(f2, argnums=1), static_argnums=(2, 3))

param = tc.backend.ones([nlayer, n])


def benchmark(f, tries=3):
    time0 = time.time()
    key = tc.backend.get_random_state(42)
    print(f(key, param, n, nlayer)[0])
    time1 = time.time()
    for _ in range(tries):
        print(f(key, param, n, nlayer)[0])
    time2 = time.time()
    print(
        "staging time: ",
        time1 - time0,
        "running time: ",
        (time2 - time1) / tries,
    )


print("without layerwise slicing jit")
benchmark(vagf1)
print("=============================")
print("with layerwise slicing jit")
benchmark(vagf2)

# 10*4: jax*T4: 235/0.36 vs. 26/0.04



"""
QML task on noisy PQC with vmapped Monte Carlo noise simulation
"""

import os

os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"
# one need this for jax+gpu combination in some cases
import time
import tensorflow as tf
import numpy as np
import optax
import tensorcircuit as tc

K = tc.set_backend("jax")

# numpy data
(x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
x_train = x_train[..., np.newaxis] / 255.0


def filter_pair(x, y, a, b):
    keep = (y == a) | (y == b)
    x, y = x[keep], y[keep]
    y = y == a
    return x, y


datapoints = 200
batch = 32
logfile = "qml_param_v2.npy"
n = 9
m = 4
maxiter = 5000
data_preparation = "v1"

x_train, y_train = filter_pair(x_train, y_train, 0, 1)

if data_preparation == "v1":
    x_train = tf.image.resize(x_train, (int(np.sqrt(n)), int(np.sqrt(n)))).numpy()
    x_train = np.array(x_train > 0.5, dtype=np.float32)
    x_train = np.squeeze(x_train).reshape([-1, n])

else:  # "v2"
    from sklearn.decomposition import PCA

    x_train = PCA(n).fit_transform(x_train.reshape([-1, 28 * 28]))


mnist_data = (
    tf.data.Dataset.from_tensor_slices((x_train[:datapoints], y_train[:datapoints]))
    .repeat(maxiter)
    .shuffle(datapoints)
    .batch(batch)
)


def f(param, seed, x, pn):
    c = tc.Circuit(n)
    px, py, pz = pn, pn, pn
    for i in range(n):
        if data_preparation == "v1":
            c.rx(i, theta=x[i] * np.pi / 2)
        else:
            c.rx(i, theta=K.atan(x[i]))
    for j in range(m):
        for i in range(n - 1):
            c.cx(i, i + 1)
            c.depolarizing(i, px=px, py=py, pz=pz, status=seed[j, i, 0])
            c.depolarizing(i + 1, px=px, py=py, pz=pz, status=seed[j, i, 1])
        for i in range(n):
            c.rz(i, theta=param[j, i, 0])
        for i in range(n):
            c.rx(i, theta=param[j, i, 1])

    ypreds = K.convert_to_tensor([K.real(c.expectation_ps(z=[i])) for i in range(n)])

    return K.mean(ypreds)


vf = tc.utils.append(K.vmap(f, vectorized_argnums=1), K.mean)


def loss(param, scale, seeds, x, y, pn):
    ypred = vf(param, seeds, x, pn)
    ypred = K.sigmoid(scale * ypred)
    y = K.cast(y, "float32")
    return (
        K.real(-y * K.log(ypred) - (1 - y) * K.log(1 - ypred)),
        ypred,
    )


def acc(yps, ys):
    yps = K.numpy(yps)
    ys = K.numpy(ys)
    yps = (np.sign(yps - 0.5) + 1) / 2
    return 1 - np.mean(np.logical_xor(ys, yps))


vgloss = K.jit(K.vvag(loss, argnums=(0, 1), vectorized_argnums=(3, 4), has_aux=True))
vloss = K.jit(K.vmap(loss, vectorized_argnums=(3, 4)))


def train(param=None, scale=None, noise=0, noc=1, fixed=True, val_step=40):
    times = []
    val_times = []
    if param is None:
        param = K.implicit_randn([m, n, 2])
    if scale is None:
        scale = 15.0 * K.ones([], dtype="float32")
    else:
        scale *= K.ones([], dtype="float32")
    if K.name == "jax":
        opt = K.optimizer(optax.adam(1e-2))
        opt2 = K.optimizer(optax.adam(5e-2))
    else:
        opt = K.optimizer(tf.keras.optimizers.Adam(1e-2))
        opt2 = K.optimizer(tf.keras.optimizers.Adam(5e-2))
    pn = noise * K.ones([], dtype="float32")

    try:
        for i, (xs, ys) in zip(range(maxiter), mnist_data):  # using tf data loader here
            xs, ys = tc.array_to_tensor(xs.numpy(), ys.numpy())
            seeds = K.implicit_randu([noc, m, n, 2])
            time0 = time.time()
            _, grads = vgloss(param, scale, seeds, xs, ys, pn)
            time1 = time.time()
            times.append(time1 - time0)
            param = opt.update(grads[0], param)
            if fixed is False:
                scale = opt2.update(grads[1], scale)
            if i % val_step == 0:
                print("%s round" % str(i))
                print("scale: ", K.numpy(scale))
                time0 = time.time()
                inference(param)
                time1 = time.time()
                val_times.append(time1 - time0)
                if len(times) > 1:
                    print("batch running time est.: ", np.mean(times[1:]))
                    print("batch staging time est.: ", times[0])
                if len(val_times) > 1:
                    print("full set running time est.: ", np.mean(val_times[1:]))
                    print("full set staging time est.: ", val_times[0])
    except KeyboardInterrupt:
        pass

    np.save(logfile, K.numpy(param))
    return param


def inference(param=None, scale=None, noise=0, noc=1, debug=False):
    pn = noise * K.ones([], dtype="float32")
    if param is None:
        param = np.load(logfile)
    if scale is None:
        scale = 15.0 * K.ones([], dtype="float32")
    else:
        scale *= K.ones([], dtype="float32")
    seeds = K.implicit_randu([noc, m, n, 2])
    vs, yps = vloss(
        param,
        scale,
        seeds,
        tc.array_to_tensor(x_train[:datapoints]),
        tc.array_to_tensor(y_train[:datapoints]),
        pn,
    )
    if debug:
        print(yps)
    print("loss: ", K.mean(vs))
    print("acc on training: %.4f" % acc(yps, y_train[:datapoints]))


if __name__ == "__main__":
    train(noise=0.005, scale=30, noc=100, fixed=False)
    # inference(noise=0.01, noc=1000, scale=40, debug=True)
